/* ***** BEGIN LICENSE BLOCK *****
   - Version: MPL 1.1/GPL 2.0/LGPL 2.1
   -
   - The contents of this file are subject to the Mozilla Public License Version
   - 1.1 (the "License"); you may not use this file except in compliance with
   - the License. You may obtain a copy of the License at
   - http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS IS" basis,
   - WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
   - for the specific language governing rights and limitations under the
   - License.
   -
   - The Original Code is CheckYesss.
   -
   - The Initial Developer of the Original Code is
   - Thomas Reitmayr <treitmayr@yahoo.com>.
   - Portions created by the Initial Developer are Copyright (C) 2005,2006
   - the Initial Developer. All Rights Reserved.
   -
   - Contributor(s):
   - -
   -
   - Alternatively, the contents of this file may be used under the terms of
   - either the GNU General Public License Version 2 or later (the "GPL"), or
   - the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
   - in which case the provisions of the GPL or the LGPL are applicable instead
   - of those above. If you wish to allow use of your version of this file only
   - under the terms of either the GPL or the LGPL, and not to allow others to
   - use your version of this file under the terms of the MPL, indicate your
   - decision by deleting the provisions above and replace them with the notice
   - and other provisions required by the LGPL or the GPL. If you do not delete
   - the provisions above, a recipient may use your version of this file under
   - the terms of any one of the MPL, the GPL or the LGPL.
   -
   - ***** END LICENSE BLOCK ***** */

const CLASS_ID = Components.ID("{ced93740-5608-11da-8cd6-0800200c9a66}");
const CLASS_NAME = "CheckYesss Settings JS Component";
const CONTRACT_ID = "@mozdev.org/checkyesss/settings_service;1";


/***************************************************************************/

/**
/* ArrayEnumerator courtesy of Brendan Eich <brendan@mozilla.org>
 */
function ArrayEnumerator(array) {
    this.array = array;
    this.index = 0;
}
ArrayEnumerator.prototype = {
    hasMoreElements: function() { return this.index < this.array.length; },
    getNext: function () { return (this.index < this.array.length) ?
                       this.array[this.index++] : null; }
};

/**
/* StringEnumerator
 */
function StringEnumerator(array) {
    this.array = array;
    this.index = 0;
}
StringEnumerator.prototype = {
    hasMore: function() { return this.index < this.array.length; },
    getNext: function () { return (this.index < this.array.length) ?
                       this.array[this.index++] : null; }
};

/***************************************************************************/

/**
 * The CheckYesss Settings Service.
 */
function CheckyesssSettingsServiceClass() {
  // allow unwrapping JS component - not needed here
  //this.wrappedJSObject = this;
  this._start();
}


CheckyesssSettingsServiceClass.prototype = {
  
  _settingsXmlEmpty: "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<settings/>\n",
  
  _globalSettings: null,
  
  _globalSettingsNames: [
    "update-interval",
    "show-whatsnew",
    "autohide-all",
    "country",
    "loglevel",
    "linksmenu-tab",
    "show-divider",
    "store-calls-dir"],
  
  _globalSettingsDefaults: [
    "update-interval",      12,
    "show-whatsnew",        true,
    "autohide-all",         false,
    "country",              "Austria",
    //"loglevel",             Components.interfaces.cyICheckyesssUtil.LOGLEVEL_WARNING,
    "loglevel",             2,
    "linksmenu-tab",        true,
    "show-divider",         true ],
  
  _accountSettings: null,
  
  _accountSettingsNames: [
    "id",                                 // unique account id
    "enabled",                            // account enabled
    "user",                               // user id
    "pw-handling",                        // how to handle the password (store/ask)
    "initial",                            // initial letter
    "provider",                           // phone provider
    "rateplan",                           // the rate plan ID
    "position",                           // position in status bar
    "show-label",                         // (auto-)show credit information
    "show-age",                           // show the age of the credit info
    "animate",                            // animated icon enabled
    "icon-theme",                         // icon theme
    "calls-autodl",                       // audo-download calls enabled
    "period-reset",                       // how to reset a balance period
    "lclick-dest",
    "mclick-dest",
    "lclick-tab",
    "mclick-tab",
    "lclick-url",
    "mclick-url"] ,
  
  _accountSettingsDefaults: [
    "enabled",              false,
    "user",                 "",
    "pw-handling",          "store",
    "initial",              "",
    "rateplan",             "auto",
    "position",             0,
    "show-label",           "yes",
    "show-age",             true,
    "animate",              true,
    "icon-theme",           "default",
    "calls-autodl",         true,
    "period-reset",         "manual",
    "lclick-dest",          "1",
    "mclick-dest",          "project",
    "lclick-tab",           true,
    "mclick-tab",           true,
    "lclick-url",           "",
    "mclick-url",           "" ],
  
  _iconsSettingsUrl: "chrome://checkyesss/skin/icons.xml",
  
  _iconSettings: null,
  
  _observerTopic: "checkyesss:settings",
  
  _settingsFile: null,
  
  _obs: null,
  _utils: null,
  
  
  /********************************************************************
  /* 
  /**/
  
  get settingsFile() {
    return this._settingsFile;
  },
  
  
  getGlobalSettings: function() {
    return this._globalSettings;
  },
  
  
  getAccountSettings: function(aId) {
    if (aId in this._accountSettings)
      return this._accountSettings[aId];
    else
      return null;
  },
  
  
  getAccountIds: function() {
    var idarr = new Array();
    for (var n in this._accountSettings)
      idarr.push(n);
    return new StringEnumerator(idarr);
  },
  
  
  getAccountDefaults: function() {
    var account = Components.classes["@mozdev.org/checkyesss/container;1"].
                    createInstance(Components.interfaces.cyICheckyesssContainer);
    this._applyDefaults(account, this._accountSettingsDefaults);
    return account;
  },
  
  
  writeNewSettings: function(aGlobal, aAccounts) {
    this._globalSettings = aGlobal;
    this._accountSettings = new Array();
    
    while (aAccounts.hasMoreElements()) {
      var account = aAccounts.getNext().
                      QueryInterface(Components.interfaces.cyICheckyesssContainer);
      if (account.hasProperty("id")) {
        var id = account.getPropertyAsString("id");
        this._accountSettings[id] = account;
      }
      else
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssSettings.writeNewSettings",
                                  "Missing property 'id' from account!",
                                  false);
    }
    
    this._writeSettings();
    this._obs.notifyObservers(this, this._observerTopic, "load");
  },
  
  
  writeModifiedSettings: function(aNotify) {
    this._writeSettings();
    if (aNotify)
      this._obs.notifyObservers(this, this._observerTopic, "load");
  },
  
  
  /********************************************************************
  /* 
  /**/
  
  _start: function() {
    
    this._initSettingsPart();
    
    // set components
    this._utils = Components.classes["@mozdev.org/checkyesss/util_service;1"].
                    getService(Components.interfaces.cyICheckyesssUtil);
    
    this._obs = Components.classes["@mozilla.org/observer-service;1"].
                  getService(Components.interfaces.nsIObserverService);
    
    //add shutdown observer
    this._obs.addObserver(this, "xpcom-shutdown", false);
    
    this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                              "Checkyesss Settings Service started", "", false);
    
    this._readSettings();
    this._readIconSettings();
  },
  
  
  /********************************************************************
  /* 
  /**/
  
  _stop: function() {
    // clean up observers
    try {
      this._obs.removeObserver(this, "xpcom-shutdown");
    }
    catch(e) {}
    
    //destroy variables
    this._settingsFile = null;
    this._globalSettings = null;
    this._accountSettings = null;
    this._obs = null;
    this._utils = null;
  },
  
  
  /********************************************************************
  /* Initialize the settings variables
  /**/
  
  _initSettingsPart: function() {
    var dirs = Components.classes["@mozilla.org/file/directory_service;1"].
                 getService(Components.interfaces.nsIProperties);
    var folder = dirs.get("ProfD", Components.interfaces.nsIFile);
    
    // append extension-specific subdirectory name
    folder.append("checkyesss");
    if (!folder.exists())
      folder.create(folder.DIRECTORY_TYPE, 0700);
    
    var file = folder.clone();
    file.append("settings.xml");
    this._settingsFile = file;
  },
  
  /********************************************************************
  /* Apply default settings to uninitialized properties.
  /**/
  
  _applyDefaults: function(aContainer, aDefaultsArray) {
    var modified = false;
    
    for (var i = 0; i+1 < aDefaultsArray.length; i += 2) {
      var name  = aDefaultsArray[i];
      if (!aContainer.hasProperty(name)) {
        var value = aDefaultsArray[i+1];
        switch (typeof(value)) {
          case "string":
            aContainer.setPropertyAsString(name, value);
            break;
          case "number":
            aContainer.setPropertyAsInt(name, value);
            break;
          case "boolean":
            aContainer.setPropertyAsBool(name, value);
            break;
          default:
            // should never be reached!
            break;
        }
        modified = true;
      }
    }
    return modified;
  },
  
  
  /********************************************************************
  /* 
  /**/
  
  _readSettings: function() {
    var writeback = false;
    var linkmap;
    var value;
    
    /**** read global settings ****/
    var settings = Components.classes["@mozdev.org/checkyesss/container;1"].
                     createInstance(Components.interfaces.cyICheckyesssContainer);
    var acpatch = Components.classes["@mozdev.org/checkyesss/container;1"].
                    createInstance(Components.interfaces.cyICheckyesssContainer);
    
    /*************/
    if (this._utils.isUpdate) {
      // Convert from pre-0.4.9-settings to current way of storing things!
      // This may be removed after some time...
      if (this._utils.compareVersions(this._utils.oldVersion, "0.4.9") < 0) {
        var pbs = Components.classes["@mozilla.org/preferences-service;1"].
                    getService(Components.interfaces.nsIPrefService);
        var branch = pbs.getBranch("checkyesss.");
        
        value = this._utils.readIntPref(branch, "interval", 4, 24, 12);
        settings.setPropertyAsInt("update-interval", value);
        
        value = this._utils.readIntPref(branch, "loglevel", 0, 2, 1);
        settings.setPropertyAsInt("loglevel", value + 1);
        
        value = this._utils.readCharPref(branch, "callsdir", "");
        settings.setPropertyAsString("store-calls-dir", value);
        
        linkmap = [ "none", "0", "1", "project", "user" ];
        value = this._utils.readIntPref(branch, "lclickdest", 0, 4, 2);
        acpatch.setPropertyAsString("lclick-dest", linkmap[value]);
        value = this._utils.readIntPref(branch, "mclickdest", 0, 4, 2);
        acpatch.setPropertyAsString("mclick-dest", linkmap[value]);
        value = this._utils.readBoolPref(branch, "lclicktab", true);
        acpatch.setPropertyAsBool("lclick-tab", value);
        value = this._utils.readBoolPref(branch, "mclicktab", true);
        acpatch.setPropertyAsBool("mclick-tab", value);
        value = this._utils.readCharPref(branch, "lclickurl", "");
        acpatch.setPropertyAsString("lclick-url", value);
        value = this._utils.readCharPref(branch, "mclickurl", "");
        acpatch.setPropertyAsString("mclick-url", value);
        acpatch.setPropertyAsString("rateplan", "auto");
        
        writeback = true;                 // store immediately
      }
    }
    // end of conversion code
    /*************/
    
    // load stored settings from file
    var doc = this._utils.readXMLDoc(this._settingsFile);
    var top = (doc) ? doc.documentElement : null;
    var globalnode;
    if (doc && top) {
      globalnode = doc.evaluate("/settings/global",
                                top,
                                null,
                                Components.interfaces.nsIDOMXPathResult.FIRST_ORDERED_NODE_TYPE,
                                null).singleNodeValue;
      
      if (globalnode)
        settings.setPropertiesFromAttributes(globalnode);
      else
        writeback = true;
    }
    else
      writeback = true;
    
    // Filter with supported settings
    //if (this._utils.keepProperties(new StringEnumerator(this._globalSettingsNames)))
    //  writeback = true;
    
    // Set the default values
    if (this._applyDefaults(settings, this._globalSettingsDefaults))
      writeback = true;
    
    // this is the final result of the global settings
    this._globalSettings = settings;
    
    /*************/
    if (this._utils.isUpdate) {
      // Convert from 0.4.9.20051212
      // this was a development version, so remove this after some time!
      if (this._utils.compareVersions(this._utils.oldVersion, "0.4.9.20051212") == 0) {
        linkmap = [ "none", "0", "1", "project", "user" ];
        value = settings.getPropertyAsInt("lclick-dest");
        acpatch.setPropertyAsString("lclick-dest", linkmap[value]);
        value = settings.getPropertyAsInt("mclick-dest");
        acpatch.setPropertyAsString("mclick-dest", linkmap[value]);
        value = settings.getPropertyAsBool("lclick-tab");
        acpatch.setPropertyAsBool("lclick-tab", value);
        value = settings.getPropertyAsBool("mclick-tab");
        acpatch.setPropertyAsBool("mclick-tab", value);
        value = settings.getPropertyAsString("lclick-url");
        acpatch.setPropertyAsString("lclick-url", value);
        value = settings.getPropertyAsString("mclick-url");
        acpatch.setPropertyAsString("mclick-url", value);
        acpatch.setPropertyAsString("rateplan", "auto");
        writeback = true;                 // store immediately
      }
    }
    // end of conversion code
    /*************/
    
    /**** read settings for each account ****/
    this._accountSettings = new Array();
    
    if (doc && top) {
      var accnodes = doc.evaluate("/settings/account",
                                  top,
                                  null,
                                  Components.interfaces.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE,
                                  null);
      
      var accnode = accnodes.iterateNext();
      while (accnode) {
        account = acpatch.clone();
        account.setPropertiesFromAttributes(accnode);
        
        if (account.hasProperty("id")) {
          if (this._applyDefaults(account, this._accountSettingsDefaults))
            writeback = true;
          
          // some general checks
          var enabled = account.getPropertyAsBool("enabled");
          var isOk = true;
          if (enabled) {
            if (!account.hasProperty("provider") ||
                !account.hasProperty("rateplan")) {
            isOk = false;
            this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                      "_readSettings",
                                      "Missing account's provider or rate plan!",
                                      false);
            }
            if (!account.hasProperty("position")) {
              isOk = false;
              this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                        "_readSettings",
                                        "Missing account's position",
                                        false);
            }
            
            // do not allow this funky setting
            if (account.getPropertyAsString("show-label") == "autoShow" &&
                settings.getPropertyAsBool("autohide-all"))
              account.setPropertyAsString("show-label", "yes");
            
            // check for double positions
            if (isOk) {
              var pos = account.getPropertyAsInt("position");
              for (var n in this._accountSettings) {
                if (pos == this._accountSettings[n].getPropertyAsInt("position")) {
                  isOk = false;
                  break;
                }
              }
              if (!isOk)
                this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                          "_readSettings",
                                          "Different accounts have identical position!",
                                          false);
            }
            if (!isOk) {
              // missing important properties -> disable this account
              account.setPropertyAsBool("enabled", false);
              enabled = false;
              writeback = true;
            }
          }
          
          this._accountSettings[account.getPropertyAsString("id")] = account;
          //dump("loaded account "+account.getPropertyAsString("id")+": (success="+isOk+", enabled="+enabled+")\n");
        }
        accnode = accnodes.iterateNext();
      }
      accnodes = null;
      accnode = null;
    }
    
    if (writeback)
      this._writeSettings();
  },
  
  
  /********************************************************************
   *
   */
  
  _writeSettings: function() {
    var parser = Components.classes["@mozilla.org/xmlextras/domparser;1"].
                   createInstance(Components.interfaces.nsIDOMParser);
    
    var doc = parser.parseFromString(this._settingsXmlEmpty, "text/xml");
    var top = doc.documentElement;
    
    this._applyDefaults(this._globalSettings, this._globalSettingsDefaults);
    var globalnode = doc.createElement("global");
    var settings = this._globalSettings.clone();
    settings.keepProperties(new StringEnumerator(this._globalSettingsNames));
    settings.getPropertiesAsAttributes(globalnode);
    top.appendChild(globalnode);
    
    for (var id in this._accountSettings) {
      this._applyDefaults(this._accountSettings[id], this._accountSettingsDefaults);
      var accnode = doc.createElement("account");
      var account = this._accountSettings[id].clone();
      account.keepProperties(new StringEnumerator(this._accountSettingsNames));
      account.getPropertiesAsAttributes(accnode);
      top.appendChild(accnode);
    }
    
    this._utils.writeXMLDoc(this._settingsFile, doc);
  },
  
  /********************************************************************
  /* 
  /**/
  
  _readIconSettings: function() {
    this._iconSettings = new Array();
    
    var ok = true;
    var request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].
                    createInstance(Components.interfaces.nsIXMLHttpRequest);
    request.open("GET", this._iconsSettingsUrl, false);
    try {
      request.send(null);
    }
    catch (e) {
      ok = false;
      this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                "cyCheckyesssSettings._readIconSettings",
                                "Error reading icons file " +
                                this._iconsSettingsUrl + ": " + e.toString(),
                                false);
    }
    
    if (ok) {
      if (request.responseXML) {
        var iconnodes = request.responseXML.evaluate("/icons/icon",
                          request.responseXML.documentElement,
                          null,
                          Components.interfaces.nsIDOMXPathResult.ORDERED_NODE_ITERATOR_TYPE,
                          null);
        
        var names = "";
        var iconnode = iconnodes.iterateNext();
        while (iconnode) {
          var icon =  Components.classes["@mozdev.org/checkyesss/container;1"].
                        createInstance(Components.interfaces.cyICheckyesssContainer);
          icon.setPropertiesFromAttributes(iconnode);
          if (icon.hasProperty("id")) {
            var id = icon.getPropertyAsString("id");
            this._iconSettings[" "+id] = icon;
            names += id + " ";
          }
          iconnode = iconnodes.iterateNext();
        }
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_INFORMATIONAL,
                                  "cyCheckyesssSettings._readIconSettings",
                                  "Available icons: " + names,
                                  false);
      }
      else {
        this._utils.writeLogEntry(Components.interfaces.cyICheckyesssUtil.LOGLEVEL_ERROR,
                                  "cyCheckyesssSettings._readIconSettings",
                                  "Error parsing icons file " +
                                  this._iconsSettingsUrl + ":\n" +
                                  request.responseText,
                                  false);
      }
    }
  },
  
  
  getIconNames: function() {
    var names = new Array();
    if (this._iconSettings) {
      for (var n in this._iconSettings)
        names.push(n.substr(1, n.length-1));
    }
    return new StringEnumerator(names);
  },
  
  
  getIconProperties: function(aName) {
    if (" "+aName in this._iconSettings)
      return this._iconSettings[" "+aName];
    return null;
  },
  
  
  /********************************************************************
  /* 
  /**/
  
  observe: function(aSubject, aTopic, aData) {
    if (aTopic == "xpcom-shutdown")
      this._stop();
  },
  
  
  /**********************************************************************/
  /*          X P C O M - S P E C I F I C   M E T H O D S               */
  /**********************************************************************/

  QueryInterface: function (aIID) {
    if (!aIID.equals(Components.interfaces.cyICheckyesssSettings) &&
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }
  
};


/******************************************************************************
 * XPCOM Functions for construction and registration
 ******************************************************************************/

var gModule = {
  _firstTime: true,
  
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType) {
    if (this._firstTime) {
      this._firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID,
                                     CLASS_NAME,
                                     CONTRACT_ID,
                                     aFileSpec,
                                     aLocation,
                                     aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType) {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID) {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return gFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) {
    return true;
  }
};

var gFactory = {
  _service: null,
  
  createInstance: function (aOuter, aIID) {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    
    if (!aIID.equals(Components.interfaces.cyICheckyesssSettings) &&
        !aIID.equals(Components.interfaces.nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    
    if (this._service == null)
      this._service = new CheckyesssSettingsServiceClass();

    return this._service.QueryInterface(aIID);
  }
};

function NSGetModule(aCompMgr, aFileSpec) {
  return gModule;
}

